from . import bruker
from . import coloring
from . import event_processing
from . import image_processing
from . import interactive_visuals
from . import io_tools
from . import reorganization
from . import static_visuals
from . import trace_processing
from . import validators
from .version import version


name = "CalSciPy"
